
/*
 * multidico
 * Copyright (C) 1992 John Heidemann <johnh@cs.ucla.edu>
 * Santa Monica, California
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

#include <stdio.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include "screenhack.h"


static GC draw_gc, erase_gc;
static unsigned int fg_gc, bg_gc;
static XFontStruct *font;
static int maxx, maxy;
static int delay;
static int axisDelay, spinDelay;
static Colormap cmap;

int show_axes_p;
int debug;

struct dpoint { double x, y; };

#define max_d 8
#define max_mat (max_d+1)
struct vector { double x[max_mat]; };
struct matrix { double x[max_mat][max_mat]; };
struct line_segment { int from, to; };


/*
 * Data storage.
 */
int num_d;
int num_mat;

/* there are C(num_d,2) planes */
#define max_planes 40
int num_planes;
XPoint *rotation_planes;
double rotations[max_planes];
double d_rotations[max_planes];   /* change in rotation */
double dd_rotations[max_planes];  /* change in change in rotation */
int cdd_rotations[max_planes];    /* how many turns to apply dd_rotations */
struct matrix Trotations[max_planes];

struct matrix Tall;

int num_points;
struct vector *points;
int num_lines;
struct line_segment *lines;

int point_set;		/* which bank of points are we using */
XPoint *xpoints[2];

int num_axis_points;
int axis_points[max_mat];



/*
 * Matrix handling & 3d transformation routines
 */

static void
MatMult (a, b, c, n)		/* c = a * b  for n x n matricies */
	struct matrix *a, *b, *c;
	int n;
{
	register int i, j, k;
	double temp;

	/* Strassen' method... what's that? */
	for (i=0; i<n; i++)		/* go through a's rows */
		for (j=0; j<n; j++) {	/* go through b's columns */
			temp = 0.0;
			for (k=0; k<n; k++)
				temp += a->x[i][k] * b->x [k][j];
			c->x[i][j] = temp;
		};	
}

static void
MatVecMult (a, b, c, n)		/* c = a * b  for a n x n, b&c 1 x n */
	struct matrix *a;
	struct vector *b, *c;
	int n;
{
	register int i, k;
	double temp;

	for (i=0; i<n; i++) {		/* go through a's rows */
		temp = 0.0;
		for (k=0; k<n; k++)
			temp += a->x[i][k] * b->x[k];
		c->x[i] = temp;
	};	
}

static void
MatCopy (a, b, n)		/* b <- a */
	struct matrix *a, *b;
	int n;
{
	register int i, j;
	for (i=0; i<n; i++)
		for (j=0; j<n; j++)
			b->x[i][j] = a->x[i][j];
}

static void
MatZero (a, n)		/* a = 0 */
	struct matrix *a;
	int n;
{
	register int i, j;
	for (i=0; i<n; i++)
		for (j=0; j<n; j++)
			a->x[i][j] = 0.0;
}

static void
MatIdent (a, n)		/* a = I */
	struct matrix *a;
	int n;
{
	register int i;
	MatZero (a, n);
	for (i=0; i<n; i++)
		a->x[i][i] = 1.0;
}



static int
randint(n)   /* 0->n-1 */
	int n;
{
	return random() % n;
}



/*
 * X stuff
 */

static void
get_geom (dpy, window)
	Display *dpy;
	Window window;
{
	XWindowAttributes xgwa;
	XGetWindowAttributes (dpy, window, &xgwa);
	cmap = xgwa.colormap;
	maxx = xgwa.width;
	maxy = xgwa.height;
}


static void
ZeroMultiCounter(c, n)
	int *c;
	int n;
{
	int i;
	for (i=0; i<n; i++)
		c[i] = 0;
}


static int
RealIncMultiCounter(c, n, place)
	int *c;
	int n, place;
{
#define BASE 2
	if (place >= n)
		return 0;
	else if (++c[place] >= BASE) {
		c[place] = 0;
		return RealIncMultiCounter(c, n, place+1);
	} else return 1;
}


static int
IncMultiCounter(c, n)
	int *c;
	int n;
{
	return RealIncMultiCounter(c, n, 0);
}


static int
figure_num_points(d)
	int d;
{
	return (int)pow(2.0, (double)d);
}

static int
figure_num_lines(d)
	int d;
{
	return d * figure_num_points(d-1);
}

static int
figure_num_planes(d)
	int d;
{
	return ((d-1) * d) / 2;
}


static void
figure_points()
{
	int i, j, k, n, d;
	int *c;

#define XtAllocArray(ELTYPE,N) ((ELTYPE*)XtMalloc((N)*sizeof(ELTYPE)))
	/*
	 * First, figure out the points.
	 */
	num_points = figure_num_points(num_d);
	/* assert(num_points <= max_points); */
	points = XtAllocArray(struct vector, num_points);
	xpoints[0] = XtAllocArray(XPoint, num_points);
	xpoints[1] = XtAllocArray(XPoint, num_points);
	c = XtAllocArray(int, num_points);   /* will be lost, sigh */
	ZeroMultiCounter(c, num_d);
	n = 0;
	do {
		for (i=0; i<num_d; i++)
			points[n].x[i] = c[i];
		n++;
	} while (IncMultiCounter(c, num_d));
	assert(num_points == n);

	/*
	 * Next connect them.
	 * We could do this more intelligently, but why bother?
	 *
	 * Connect points that differ by only one coordinate.
	 */
	num_lines = figure_num_lines(num_d);
	/* assert(num_lines <= max_lines); */
	lines = (struct line_segment*)XtMalloc(num_lines * sizeof(struct line_segment));
	for (n=i=0; i<num_points; i++)
		for (j=i+1; j<num_points; j++) {
			for (d=k=0; k<num_d; k++)
				if (points[i].x[k] != points[j].x[k])
					d++;
			if (d==1) {
				lines[n].from = i;
				lines[n].to = j;
				n++;
			};
		};
	assert(num_lines == n);

	/*
	 * Now determine the planes of rotation.
	 */
	num_planes = figure_num_planes(num_d);
	assert(num_planes <= max_planes);
	rotation_planes = XtAllocArray(XPoint, num_planes);
	for (n=i=0; i<num_d; i++)
		for (j=i+1; j<num_d; j++) {
			rotation_planes[n].x = i;
			rotation_planes[n].y = j;
			n++;
		};
	assert(num_planes == n);

	/*
	 * Potential random initial rotations.
	 */
#define FRAC (1024*16)
	if (get_boolean_resource("randomStart", "Boolean")) {
		for (i=0; i<num_d; i++)
			rotations[i] = 2.0 * randint(FRAC) * M_PI / FRAC;
	};
}


void
figure_axis_points()
{
	int i, j, num_set;

	show_axes_p = get_boolean_resource ("showAxes", "Boolean");
	for (num_axis_points=i=0; i<num_points; i++) {
		for (num_set=j=0; j<num_d; j++)
			if (points[i].x[j] != 0.0)
				num_set++;
		if (num_set <= 1)
			axis_points[num_axis_points++] = i;
	};
}


void
init_font_stuff(dpy, window)
	Display *dpy;
	Window window;
{
	char *fontname = get_string_resource ("font", "Font");
#define FONT_NAME			"-*-times-*-*-*-*-18-*-*-*-*-*-*-*"
	char **list;
	int i, foo;

	/*
	 * Font stuff from "noseguy".
	 */
	if (!fontname || !(font = XLoadQueryFont(dpy, fontname))) {
		list = XListFonts(dpy, FONT_NAME, 32767, &foo);
		for (i = 0; i < foo; i++)
			if (font = XLoadQueryFont(dpy, list[i]))
				break;
		if (!font) {
			fprintf (stderr, "%s: Can't find a large font.", progname);
			exit (1);
		}
		XFreeFontNames(list);
	}

}


void
init_x_stuff(dpy, window)
	Display *dpy;
	Window window;
{
	XGCValues gcv;
	XWindowAttributes xgwa;
	int options = 0;

	XGetWindowAttributes (dpy, window, &xgwa);
	get_geom (dpy, window);

	delay = get_integer_resource ("delay", "Integer");
	if (delay < 0) delay = 0;

	axisDelay = get_integer_resource ("axisDelay", "Integer");
	if (axisDelay < 0) axisDelay = 0;

	spinDelay = get_integer_resource ("spinDelay", "Integer");
	if (spinDelay < 0) spinDelay = 0;

	num_d = get_integer_resource ("dimension", "Integer");
	if (num_d < 2) num_d = 2;
	if (num_d >= max_d) {
		fprintf (stderr, "Dimensionality too high.  Reset to maximum of %d.\n", max_d-1);
		num_d = max_d-1;
	};
	num_mat = num_d+1;


	debug = get_boolean_resource ("debug", "Boolean");

	init_font_stuff(dpy, window);

#if 0	
	gcv.function = GXxor;
	gcv.foreground = default_fg_pixel =
		get_pixel_resource ("foreground", "Foreground", dpy, cmap);
	draw_gc = erase_gc = 
		XCreateGC (dpy, window, GCFunction|GCForeground, &gcv);
#endif
	options |= GCForeground|GCFont;
	gcv.foreground = get_pixel_resource ("foreground", "Foreground", dpy, cmap);
	gcv.font = font->fid;
	draw_gc = XCreateGC (dpy, window, GCForeground, &gcv);
	gcv.foreground = get_pixel_resource ("background", "Background", dpy, cmap);
	erase_gc = XCreateGC (dpy, window, GCForeground, &gcv);
}


void
init_multidico (dpy, window)
	Display *dpy;
	Window window;
{
	int i;

	init_x_stuff(dpy, window);

	figure_points();

	/*
	 * Fix the d+1 coord of all points.
	 */
	for (i=0; i<num_points; i++)
		points[i].x[num_d] = 1;

	figure_axis_points();
}


static void
draw_multidico(dpy, window, gc, set)
	Display *dpy;
	Window window;
	GC gc;
	int set;
{
	int i;
	int from, to;
	char tmps[80];

	for (i=0; i<num_lines; i++) {
		from = lines[i].from;
		to = lines[i].to;
		XDrawLine(dpy, window, gc,
			  xpoints[set][from].x, xpoints[set][from].y,
			  xpoints[set][to].x, xpoints[set][to].y);
	};

	if (show_axes_p)
		for (i=0; i<num_axis_points; i++) {
			sprintf (tmps, "%d", i);
			to = axis_points[i];
			XDrawString(dpy, window, gc,
				    xpoints[set][to].x, xpoints[set][to].y,
				    tmps, strlen(tmps));
		};
}

static time_t
timer_time(start_time)
	time_t *start_time;
{
	time_t current_time;

	if (*start_time == 0)
		time(start_time);
	time(&current_time);
	return current_time - *start_time;
}

static void
move_multidico()
{
#if 1
	static int inited = 0;
	int i;
	int stationary = 0;
/* NEEDSWORK: These should be resources */
#define default_cdd_rotation 10
#define default_dd_rotation (M_PI/1024.0)
	int axis, max_axis;
	static time_t axis_timer, spin_timer, debug_timer;
	int faster;

	if (timer_time(&spin_timer) > spinDelay) {
		spin_timer = 0;
		/*
		 * Change rotation?
		 *
		 * 66% chance if stationary, 33% if not.
		 */
		stationary = 1;
		for (i=0; i<num_planes; i++)
			if (d_rotations[i] != 0.0 || cdd_rotations[i]) {
				stationary = 0;
				break;
			};
		if (randint(3) < 1+stationary) {
			/* Change!  But what? */
			if (axisDelay == 0)
				max_axis = num_planes;
			else max_axis = timer_time(&axis_timer) / axisDelay + 1;
			if (max_axis > num_planes || max_axis < 0)
				max_axis = num_planes;
			axis = randint(max_axis);
			
			/*
			 * And how much?  33% chance faster, 66% slower.
			 * If stopped, slower doesn't start it moving
			 * unless we're stationary.
			 */
			cdd_rotations[axis] = default_cdd_rotation +
						randint(7) - 3;
			faster = (randint(3) < 1);
			if (faster || dd_rotations[axis] != 0.0 || stationary)
				dd_rotations[axis] = default_dd_rotation *
					(dd_rotations[axis] >= 0.0 ? 1 : -1) *
						(faster ? 1 : -1);
			if (debug)
				fprintf(stderr,"axis %d made %s at %d\n",
					axis, faster ? "faster" : "slower",
					timer_time(&debug_timer));
		};
	};
#else
	int i;
	static int inited = 0;

	if (!inited) {
		d_rotations[0] = 0.010; /* xy */
		d_rotations[1] = 0.005; /* xz */
		d_rotations[2] = 0.000; /* xw */
		d_rotations[3] = 0.000; /* yz */
		d_rotations[4] = 0.010; /* yw */
		d_rotations[5] = 0.000; /* zw */
	};
#endif

	/*
	 * Rotate.
	 */
	for (i=0; i<num_planes; i++) {
		if (cdd_rotations[i]) {
			cdd_rotations[i]--;
			d_rotations[i] += dd_rotations[i];
		};
		rotations[i] += d_rotations[i];
	};
}

#ifdef NEED_SINCOS
static void
sincos(a, s, c)
	double a, *s, *c;
{
	*s = sin(a);
	*c = cos(a);
}
#endif /* NEED_SINCOS */

static void
calc_transformation()
{
	double cosa, sina;
	int p1, p2;
	struct matrix Ttmp;
	struct matrix Tpre, Tuser, Tpretranspose, Tscale, Tposttranspose;
	int i;
	struct dpoint scale, range, offset;
	double scale_used;

	/*
	 * Adjust the data.
	 * Since the data goes from 0->1 on each axis,
	 * we shift it by -0.5 here to center the figure.
	 */
	MatIdent(&Tpre, num_mat);
	for (i=0; i<num_d; i++) {
		Tpre.x[i][num_d] = -0.5;
	};
		

	/*
	 * Figure the rotation.
	 */
	MatIdent(&Tuser, num_mat);
	for (i=0; i<num_planes; i++) {
		p1 = rotation_planes[i].x;
		p2 = rotation_planes[i].y;
		sincos(rotations[i], &sina, &cosa);
		MatIdent(&Trotations[i], num_mat);
		Trotations[i].x[p1][p1] = Trotations[i].x[p2][p2] = cosa;
		Trotations[i].x[p1][p2] = sina;
		Trotations[i].x[p2][p1] = -sina;
		MatMult(&Trotations[i], &Tuser, &Ttmp, num_mat);
		MatCopy(&Ttmp, &Tuser, num_mat);
	};

/* Now calculate the scaling matrix */
	if (1) {
		/*
		 * Calculate the best scale of the two axes.
		 * Multiply by 0.8 to use 80% of the display.
		 * Divide by the sqrt(2.0) because it's bigest when
		 * rotated by 45 degrees.
		 *
		 * This principle generalizes to sqrt(num_d).
		 */
#define border_width (0.1)
		range.x = range.y = sqrt((double) num_d);
		scale.x = (1.0-2*border_width) * maxx / range.x;
		scale.y = (1.0-2*border_width) * maxy / range.y;
		scale_used = ((scale.x < scale.y) ? scale.x : scale.y);
		offset.x = maxx / 2.0;
		offset.y = maxy / 2.0;

		/*
		 * Setup & compute the matricies
		 */
		MatIdent (&Tpretranspose, num_mat);
		Tpretranspose.x[0][num_d] = 0;
		Tpretranspose.x[1][num_d] = 0;

		MatIdent (&Tscale, num_mat);
		Tscale.x[0][0] = scale_used;
		Tscale.x[1][1] = -scale_used;

		MatIdent (&Tposttranspose, num_mat);
		Tposttranspose.x[0][num_d] = offset.x;
		Tposttranspose.x[1][num_d] = offset.y;

		MatMult (&Tscale, &Tpretranspose, &Ttmp, num_mat);
		MatMult (&Tposttranspose, &Ttmp, &Tscale, num_mat);
	} else MatIdent (&Tscale, num_mat);

	/*
	 * Put it all together.
	 */
	MatMult(&Tuser, &Tpre, &Ttmp, num_mat);
	MatMult(&Tscale, &Ttmp, &Tall, num_mat);
}


static void
translate_point(real, screen_image)
	struct vector *real;
	XPoint *screen_image;
{
	struct vector image;
	MatVecMult (&Tall, real, &image, num_mat);
	screen_image->x = (short)image.x[0];
	screen_image->y = (short)image.x[1];
}

static void
translate_points(set)
	int set;
{
	int i;
	calc_transformation();
	for (i=0; i<num_points; i++) {
		translate_point(&points[i], &xpoints[set][i]);
	};
}



static void
multidico(dpy, window)
	Display *dpy;
	Window window;
{
	static int translated;
	static int gtick = 0;
	int this_set = point_set, next_set = !point_set;

	if (gtick++ == 500)
		get_geom (dpy, window), gtick = 0;

	if (!translated) {
		translate_points(this_set);
		translated = 1;
	};
	draw_multidico(dpy, window, draw_gc, this_set);
	XSync (dpy, True);

	/* Set up next place */
	move_multidico();
	translate_points(next_set);

	/* Wait */
	if (delay) usleep (delay);
	draw_multidico(dpy, window, erase_gc, this_set);
	point_set = next_set;
}


char *progclass = "Multidico";

char *defaults [] = {
	"*background:	black",
	"*foreground:	white",
	"*dimension:	4",
	"*delay:	300000",
	"*spinDelay:	2",
	"*axisDelay:	4",
	"*showAxes:	false",
	"*randomStart:	false",
	"*debug:	false",
	"*font:		-*-new century schoolbook-*-r-*-*-*-180-*-*-*-*-*-*",
	0
};

XrmOptionDescRec options [] = {
	{ "-delay",		".delay",	XrmoptionSepArg, 0 },
	{ "-dimension",		".dimension",	XrmoptionSepArg, 0 },
	{ "-axis-delay",	".axisDelay",	XrmoptionSepArg, 0 },
	{ "-spin-delay",	".spinDelay",	XrmoptionSepArg, 0 },
	{ "-show-axes",		".showAxes",	XrmoptionNoArg, "true" },
	{ "-random-start",	".randomStart",	XrmoptionNoArg, "true" },
	{ "-zeroed-start",	".randomStart",	XrmoptionNoArg, "false" },
	{ "-no-axes",		".showAxes",	XrmoptionNoArg, "false" },
	{ "-font",		".font",	XrmoptionSepArg, 0 },
};
int options_size = (sizeof (options) / sizeof (options[0]));
	
void
screenhack (dpy, window)
	Display *dpy;
	Window window;
{
	init_multidico(dpy, window);
	for (;;)
		multidico (dpy, window);
}
